#
# Copyright 2002 Roland Stigge
#
# This file is part of ImageWorkshop
#
# ImageWorkshop is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# ImageWorkshop is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with ImageWorkshop; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#

# system modules
import os.path
import time
import math

# Tkinter
import Tkinter
import tkMessageBox

# Python Imaging Library (PIL)
import Image
import ImageDraw

# Numerical Python
import Numeric
import LinearAlgebra

# own modules
import GUI              # common dialogs, elements
import ImageGUI         # Image handling
import ImageWorkshop
import NumericImage
import Version

#
# classes for basic user interface problems
#

class AboutDialog(GUI.BigInfo):
    "The 'about' help window"
    def body(self, master):
        label = Tkinter.Label(master, text = 'ImageWorkshop\nVersion ' +
                              Version.version + '\nRoland Stigge 2002')
        label.pack()
        message = Tkinter.Message(master, text = 'This program is NOT a drawing program. (Use GIMP for that.) Its domain should be image analysis (measurement, spectrum analysis).', aspect=300)
        message.pack()

class SaveDialogJPEG(GUI.Dialog):
    "asks for Quality (1..[75]..100), optimize, progressive"

    def body(self, master):
        self.quality = Tkinter.IntVar()
        Tkinter.Scale(master, from_ = 1, to_ = 100, orient = Tkinter.HORIZONTAL, digits = 3, label = "Quality", variable = self.quality).grid(row = 0, sticky = Tkinter.W)
        self.quality.set(75)
        self.optimize = Tkinter.IntVar()
        self.optimize.set(1)
        Tkinter.Checkbutton(master, text = "optimize", variable = self.optimize).grid(row = 1, sticky = Tkinter.W)
        self.progressive = Tkinter.IntVar()
        self.progressive.set(0)
        Tkinter.Checkbutton(master, text = "progressive", variable = self.progressive).grid(row = 2, sticky = Tkinter.W)

    def apply(self):
        self.result = {"quality": self.quality.get()}
        if self.optimize.get():
            self.result["optimize"] = 1
        if self.progressive.get():
            self.result["progressive"] = 1

class FileImageWindow(ImageGUI.StaticImageWindow):
    "represents a file image"
    
    def __init__(self, master, filename):
        """
        constructor:
           master:    the parent widget (e.g. Window)
           filename:  the name of the file to display
        """
        #
        # internal representation of the Image object in self.image
        #
        self.image = Image.open(filename) # a PIL Image object
        
        # 256 gray values or RGB? -> 2 modes to distinguish
        if self.image.mode == "1":        # 1 -> L (256-gray)
            self.image = self.image.convert("L")
        elif self.image.mode != "L" and self.image.mode != "RGB":
            # color -> RGB
            self.image = self.image.convert("RGB")
        # XXX: 32 bit image support (I and F)

        ImageGUI.StaticImageWindow.__init__(self, master, os.path.basename(filename))

#
# classes for image generation
#

class SizeDialog(GUI.Dialog):
    """
    asks for width and height
    on success, self.width and self.height are not None
    cancel -> width = height = None
    * deriving classes have to notice master grid rows 0 and 1 as used
    * they have to call (in the end) body from body and validate from validate
      e.g. in method body:
               ...
               return SizeDialog.body(self, master)
    * they must set self.width = self.height = None if validate error
    """
    
    def body(self, master):
        Tkinter.Label(master, text = "Width:").grid(row = 0)  # > 0
        Tkinter.Label(master, text = "Height:").grid(row = 1) # > 0

        self.widthentry = Tkinter.Entry(master)
        self.widthentry.insert(Tkinter.END, "128")
        self.heightentry = Tkinter.Entry(master)
        self.heightentry.insert(Tkinter.END, "128")

        self.widthentry.grid(row = 0, column = 1)
        self.heightentry.grid(row = 1, column = 1)

        self.width = self.height = None

        return self.widthentry # initial focus

    def validate(self):
        try:
            self.width = int(self.widthentry.get())
        except ValueError:
            tkMessageBox.showwarning("Bad input", "Width must be an integer!")
            self.width = self.height = None
            return 0
        try:
            self.height = int(self.heightentry.get())
        except ValueError:
            tkMessageBox.showwarning("Bad input",
                                     "Height must be an integer!")
            self.width = self.height = None
            return 0
        if self.width < 1:
            tkMessageBox.showwarning("Bad input",
                                     "Width must be greater than 0!")
            self.width = self.height = None
            return 0
        if self.height < 1:
            tkMessageBox.showwarning("Bad input",
                                     "Height must be greater than 0!")
            self.width = self.height = None
            return 0
            
        return 1
    
class HomogeneousDialog(SizeDialog):
    "asks for size (SizeDialog) and color (value)"
    
    def body(self, master):
        Tkinter.Label(master, text = "Value:").grid(row = 2)  # 0 .. 255

        self.e3 = Tkinter.Entry(master)
        self.e3.insert(Tkinter.END, "128")

        self.e3.grid(row = 2, column = 1)

        return SizeDialog.body(self, master) # first focus

    def validate(self):
        if not SizeDialog.validate(self):
            return 0
        
        try:
            self.value = int(self.e3.get())
        except ValueError:
            tkMessageBox.showwarning("Bad input", "Value must be an integer!")
            self.width = self.height = None
            return 0
        if self.value < 0 or self.value > 255:
            tkMessageBox.showwarning("Bad input",
                                     "Value must be in {0, ..., 255}")
            self.width = self.height = None
            return 0
            
        return 1
    
class HomogeneousImageWindow(ImageGUI.StaticImageWindow):
    "represents a homogeneous generated image"

    def __init__(self, master, width, height, value, name = ""):

        self.image = Image.new("L", (width, height), value)
        ImageGUI.StaticImageWindow.__init__(self, master, name)

class ChessDialog(SizeDialog):
    "asks for size (SizeDialog) and frequencies"
    
    def body(self, master):
        Tkinter.Label(master, text = "x-Frequency:").grid(row = 2)  # >= 0
        Tkinter.Label(master, text = "y-Frequency:").grid(row = 3)  # >= 0

        self.e3 = Tkinter.Entry(master)
        self.e3.insert(Tkinter.END, "4") # 4 periods like chess
        self.e4 = Tkinter.Entry(master)
        self.e4.insert(Tkinter.END, "4")

        self.e3.grid(row = 2, column = 1)
        self.e4.grid(row = 3, column = 1)

        return SizeDialog.body(self, master) # first focus

    def validate(self):
        if not SizeDialog.validate(self):
            return 0
        
        try:
            self.frequencyx = float(self.e3.get())
            self.frequencyy = float(self.e4.get())
        except ValueError:
            tkMessageBox.showwarning("Bad input",
                                     "Frequencies must be float values")
            self.width = self.height = None
            return 0
        if self.frequencyx < 0 or self.frequencyy < 0:
            tkMessageBox.showwarning("Bad input",
                                     "Frequencies must not be smaller than 0")
            self.width = self.height = None
            return 0
            
        return 1
    
class ChessImageWindow(ImageGUI.StaticImageWindow):
    "represents a generated chess image"

    def __init__(self, master, width, height, frequencyx, frequencyy,
                 name = ""):

        self.image = Image.new("L", (width, height))
        periodx = width / frequencyx
        periody = height / frequencyy
        fx = lambda i, period = periodx: i % period < period / 2
        fy = lambda i, period = periody: i % period < period / 2
        for i in range(width):
            for j in range(height):
                self.image.putpixel((i, j), (fx(i) ^ fy(j)) * 255)
        ImageGUI.StaticImageWindow.__init__(self, master, name)

class GradientDialog(SizeDialog):
    "asks for size (SizeDialog) and additional paramaters (distance, angle)"

    def body(self, master):
        self.modevar = Tkinter.IntVar()
        self.modevar.set(0) # 0 = horizontal, 1 = vertical, 2 = diagonal
        Tkinter.Radiobutton(master, text = "Horizontal", value = 0,
                            variable = self.modevar).grid(row = 2,
                                                          columnspan = 2,
                                                          sticky = Tkinter.W)
        Tkinter.Radiobutton(master, text = "Vertical", value = 1,
                            variable = self.modevar).grid(row = 3,
                                                          columnspan = 2,
                                                          sticky = Tkinter.W)
        Tkinter.Radiobutton(master, text = "Diagonal", value = 2,
                            variable = self.modevar).grid(row = 4,
                                                          columnspan = 2,
                                                          sticky = Tkinter.W)

        return SizeDialog.body(self, master) # first focus

    def validate(self):
        if not SizeDialog.validate(self):
            return 0
        
        if self.width < 2 or self.height < 2:
            tkMessageBox.showwarning("Bad input",
                                     "Size must be at least 2 x 2")
            self.width = self.height = None
            return 0
            
        return 1
    
class GradientImageWindow(ImageGUI.StaticImageWindow):
    "represents a generated Dirac image"

    def __init__(self, master, width, height, mode, name = ""):

        self.image = Image.new("L", (width, height))
        if mode == 0: # horizontal
            g = lambda i, j, width = width: i * 255 / (width - 1)
        elif mode == 1: # vertical
            g = lambda i, j, height = height: j * 255 / (height - 1)
        elif mode == 2: # diagonal
            g = lambda i, j, height = height, width = width: (i + j) * 255 / (width + height - 2)
        for i in range(width):
            for j in range(height):
                self.image.putpixel((i, j), g(i,j))
                    
        ImageGUI.StaticImageWindow.__init__(self, master, name)

class DiracDialog(SizeDialog):
    "asks for size (SizeDialog) and additional paramaters (distance, angle)"

    def body(self, master):
        Tkinter.Label(master, text = "Distance:").grid(row = 2)
                                                    # 0 .. min(width, height)
        Tkinter.Label(master, text = "Angle:").grid(row = 3)  # 0 .. 90 (deg.)

        self.e3 = Tkinter.Entry(master)
        self.e3.insert(Tkinter.END, "8")
        self.e4 = Tkinter.Entry(master)
        self.e4.insert(Tkinter.END, "30")

        self.e3.grid(row = 2, column = 1)
        self.e4.grid(row = 3, column = 1)

        self.cbvar = Tkinter.IntVar()
        self.cbvar.set(1)
        self.cb = Tkinter.Checkbutton(master, text = "Symmetric",
                                      variable = self.cbvar) # 2 impulses?
        self.cb.grid(row = 4, columnspan = 2, sticky = Tkinter.W)

        return SizeDialog.body(self, master) # first focus

    def validate(self):
        if not SizeDialog.validate(self):
            return 0
        
        try:
            self.distance = int(self.e3.get())
        except ValueError:
            tkMessageBox.showwarning("Bad input",
                                     "Distance must be an integer!")
            self.width = self.height = None
            return 0
        try:
            self.angle = int(self.e4.get())
        except ValueError:
            tkMessageBox.showwarning("Bad input", "Angle must be an integer!")
            self.width = self.height = None
            return 0
        if self.distance < 0 or self.distance > min(self.width, self.height):
            tkMessageBox.showwarning("Bad input",
                                     "Value must be smaller than width and height")
            self.width = self.height = None
            return 0
        if self.angle < 0 or self.angle > 90:
            tkMessageBox.showwarning("Bad input",
                                     "Angle must be in {0, ..., 90}")
            self.width = self.height = None
            return 0
            
        return 1
    
class DiracImageWindow(ImageGUI.StaticImageWindow):
    "represents a generated Dirac image"

    def __init__(self, master, width, height, distance, angle, symmetric, name = ""):

        self.image = Image.new("L", (width, height), 0)
        self.image.putpixel((int(distance * math.cos(angle*math.pi/180)),
                             int(distance * math.sin(angle*math.pi/180))), 255)
        if symmetric:
            self.image.putpixel((width-1-int(distance *
                                             math.cos(angle*math.pi/180)),
                                 height-1-int(distance *
                                              math.sin(angle*math.pi/180))),
                                255)
        ImageGUI.StaticImageWindow.__init__(self, master, name)

class SinDialog(SizeDialog):
    "asks for needed parameters"

    def body(self, master):
        Tkinter.Label(master, text = "x: Frequency:").grid(row = 2) # >= 0
        Tkinter.Label(master, text = "x: Phase").grid(row = 3) # all the same
        Tkinter.Label(master, text = "y: Frequency:").grid(row = 4) # >= 0
        Tkinter.Label(master, text = "y: Phase").grid(row = 5) # all the same

        self.e3 = Tkinter.Entry(master)
        self.e3.insert(Tkinter.END, "1.0")
        self.e4 = Tkinter.Entry(master)
        self.e4.insert(Tkinter.END, "90.0")
        self.e5 = Tkinter.Entry(master)
        self.e5.insert(Tkinter.END, "2.0")
        self.e6 = Tkinter.Entry(master)
        self.e6.insert(Tkinter.END, "0.0")

        self.e3.grid(row = 2, column = 1)
        self.e4.grid(row = 3, column = 1)
        self.e5.grid(row = 4, column = 1)
        self.e6.grid(row = 5, column = 1)

        return SizeDialog.body(self, master) # first focus

    def validate(self):
        if not SizeDialog.validate(self):
            return 0
        
        try:
            self.frequencyx = float(self.e3.get())
            self.phasex     = float(self.e4.get())
            self.frequencyy = float(self.e5.get())
            self.phasey     = float(self.e6.get())
        except ValueError:
            tkMessageBox.showwarning("Bad input",
                                     "Frequencies and phases must be float!")
            self.width = self.height = None
            return 0
        if self.frequencyx < 0 or self.frequencyy < 0:
            tkMessageBox.showwarning("Bad input",
                                     "Frequencies must not be smaller than 0")
            self.width = self.height = None
            return 0
            
        return 1
    
class SinImageWindow(ImageGUI.StaticImageWindow):
    "represents a generated Sin/Cos image"

    def __init__(self, master, width, height, frequencyx, phasex, frequencyy, phasey, name = ""):

        self.image = Image.new("L", (width, height))
        for i in range(width):
            for j in range(height):
                self.image.putpixel((i, j),
                                    int((math.sin(i * 2 * math.pi / width *
                                                      frequencyx +
                                                  phasex * math.pi / 180) *
                                         math.sin(j * 2 * math.pi / height *
                                                      frequencyy +
                                                  phasey * math.pi / 180) +
                                         1) * 255 / 2))
        ImageGUI.StaticImageWindow.__init__(self, master, name)

class GaussDialog(SizeDialog):
    "asks for needed parameters"

    def body(self, master):
        Tkinter.Label(master, text = "Standard deviation x:").grid(row = 2)
                                                                # > 0
        Tkinter.Label(master, text = "Standard deviation y:").grid(row = 3)
                                                                # > 0
        self.e3 = Tkinter.Entry(master)
        self.e3.insert(Tkinter.END, "8")
        self.e4 = Tkinter.Entry(master)
        self.e4.insert(Tkinter.END, "8")

        self.e3.grid(row = 2, column = 1)
        self.e4.grid(row = 3, column = 1)

        self.cbvarx = Tkinter.IntVar()
        self.cbvarx.set(1)
        self.cbx = Tkinter.Checkbutton(master, text = "Gauss at x-axis",
                                       variable = self.cbvarx) 
        self.cbx.grid(row = 4, columnspan = 2, sticky = Tkinter.W)

        self.cbvary = Tkinter.IntVar()
        self.cbvary.set(1)
        self.cby = Tkinter.Checkbutton(master, text = "Gauss at y-axis",
                                       variable = self.cbvary) 
        self.cby.grid(row = 5, columnspan = 2, sticky = Tkinter.W)

        return SizeDialog.body(self, master) # first focus

    def validate(self):
        if not SizeDialog.validate(self):
            return 0
        
        try:
            self.stddevx = float(self.e3.get())
            self.stddevy = float(self.e4.get())
        except ValueError:
            tkMessageBox.showwarning("Bad input",
                                     "Standard deviations must be float!")
            self.width = self.height = None
            return 0
        if self.stddevx <= 0 or self.stddevy <= 0:
            tkMessageBox.showwarning("Bad input",
                                     "Standard deviations must be greater than 0")
            self.width = self.height = None
            return 0
            
        return 1
    
class GaussImageWindow(ImageGUI.StaticImageWindow):
    "represents a generated Gauss image"

    def __init__(self, master, width, height, gaussx, stddevx, gaussy, stddevy, name = ""):

        self.image = Image.new("L", (width, height))
        if gaussx:
            gx = lambda x, stddev = stddevx: math.exp(- abs(float(x) / stddev) ** 2 / 2)
        else:
            gx = lambda x: 1
        if gaussy:
            gy = lambda y, stddev = stddevy: math.exp(- abs(float(y) / stddev) ** 2 / 2)
        else:
            gy = lambda y: 1
        for i in range(width):
            for j in range(height):
                self.image.putpixel((i, j), gx(i - width / 2) * gy(j - height / 2) * 255)

        ImageGUI.StaticImageWindow.__init__(self, master, name)

class SiemensDialog(GUI.Dialog):
    """
    asks for size and number of Siemens star edges
    on success, self.width != None
    results in self.width, self.edges
    """
    
    def body(self, master):
        "build the main dialog fields"
        
        Tkinter.Label(master, text = "Width (=Height):").grid(row = 0)  # > 0
        Tkinter.Label(master, text = "Edges:").grid(row = 1)  # > 0

        self.widthentry = Tkinter.Entry(master)
        self.widthentry.insert(Tkinter.END, "256")
        self.widthentry.grid(row = 0, column = 1)

        self.edgesentry = Tkinter.Entry(master)
        self.edgesentry.insert(Tkinter.END, "32")
        self.edgesentry.grid(row = 1, column = 1)

        return self.widthentry # first focus

    def validate(self):
        "check values"
        
        try:
            self.width = int(self.widthentry.get())
            self.edges = int(self.edgesentry.get())
        except ValueError:
            tkMessageBox.showwarning("Bad input", "Values must be integers!")
            self.width = self.edges = None
            return 0
        if self.width < 1 or self.edges < 1:
            tkMessageBox.showwarning("Bad input",
                                     "Values must be > 0")
            self.width = self.edges = None
            return 0
            
        return 1

class DrawExt(ImageDraw.Draw):
    """
    unfortunately, PIL ImageDraw.Draw pieslice() (like its arc-related
    siblings) doesn't support the float type for start and end.
    We need some more precision here, so we reimplement it directly in Python
    """
    def pieslice(self, (x1, y1, x2, y2), start, end,
                 fill = None, outline = None):
        while end < start:
            end += 360
        centerx = float(x1 + x2) / 2
        centery = float(y1 + y2) / 2
        radiusx = x2 - centerx
        radiusy = y2 - centery
        xy = [(centerx, centery)] # list of tuples to build for polygon
        while start < end:
            x = math.cos(start * math.pi / 180) * radiusx + centerx
            y = math.sin(start * math.pi / 180) * radiusy + centery
            xy.append((x, y))
            start += 1
        x = math.cos(end * math.pi / 180) * radiusx + centerx
        y = math.sin(end * math.pi / 180) * radiusy + centery
        xy.append((x, y))
        self.polygon(xy, fill = fill, outline = outline)
    
class SiemensImageWindow(ImageGUI.StaticImageWindow):
    "represents a generated Siemens star image"

    def __init__(self, master, width, edges, name = ""):

        self.image = Image.new("L", (width, width), 255)
        for i in range(edges):
            draw = DrawExt(self.image)
            draw.pieslice((0, 0, width, width),
                          float(i * 2) * 360 / (edges * 2),
                          float(i * 2 + 1) * 360 / (edges * 2),
                          outline = 0, fill = 0)
        ImageGUI.StaticImageWindow.__init__(self, master, name)

