#
# Copyright 2002 Roland Stigge
#
# This file is part of ImageWorkshop
#
# ImageWorkshop is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# ImageWorkshop is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with ImageWorkshop; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#

# Tkinter
import Tkinter

# Python Imaging Library (PIL)
import Image

# Numerical Python
import Numeric

# own modules
import GUI              # common dialogs, elements
import ImageGUI         # Image handling
import NumericImage

#
# classes for image data calculation
#

class ScalarsWindow(Tkinter.Toplevel):
    "Shows (many) scalar Values about an ImageWindow"
    def __init__(self, master):
        "master: ImageWindow instance"
        Tkinter.Toplevel.__init__(self, master)
        self.title(master.name + " Scalar Values")
        
        channels = master.getnumericimage()
        for i in range(len(channels)):
            channels[i] = channels[i].astype(Numeric.Float)
        #image = master.getview()
        
        f = Tkinter.Frame(self)
        f.pack(padx=5, pady=5)

        Tkinter.Label(f, text = "Size", anchor = Tkinter.W).grid(row = 0, sticky = Tkinter.W)
        Tkinter.Label(f, text = "Width", anchor = Tkinter.W).grid(row = 1, sticky = Tkinter.W)
        Tkinter.Label(f, text = "Height", anchor = Tkinter.W).grid(row = 2, sticky = Tkinter.W)
        Tkinter.Label(f, text = str(master.width() * master.height()), anchor = Tkinter.W).grid(row = 0, column = 1, sticky = Tkinter.E)
        Tkinter.Label(f, text = str(master.width()), anchor = Tkinter.W).grid(row = 1, column = 1, sticky = Tkinter.E)
        Tkinter.Label(f, text = str(master.height()), anchor = Tkinter.W).grid(row = 2, column = 1, sticky = Tkinter.E)

        f = Tkinter.Frame(self)
        f.pack(padx=5, pady=5)

        computations = (
            ("Channel", lambda i: str(i)),
            ("Number of possible values", lambda i, c = channels: c[i]._get_G()),
            ("Number of different values", lambda i, c = channels: c[i].numberofdifferentvalues()),
            ("Minimum value", lambda i, c = channels: c[i].min()),
            ("Maximum value", lambda i, c = channels: c[i].max()),
            ("Grey tone span", lambda i, c = channels: c[i].grayspan()),
            ("Dynamic", lambda i, c = channels: "%.2f" % c[i].dynamic()),
            ("Modal value", lambda i, c = channels: c[i].modal()),
            ("Median", lambda i, c = channels: c[i].median()),
            ("Mean value (average)", lambda i, c = channels: "%.2f" % c[i].average()),
            ("Variance", lambda i, c = channels: "%.2f" % c[i].variance()),
            ("Standard deviation", lambda i, c = channels: "%.2f" % abs(c[i].stddev())),
            #("Asymmetry", lambda i, c = channels: "%.2f" % c[i].asymmetry()),
            ("Asymmetry, normalized", lambda i, c = channels: "%.2f" % c[i].norm_asymmetry()),
            #("Kurtosis", lambda i, c = channels: "%.2f" % c[i].kurtosis()),
            ("Kurtosis, normalized", lambda i, c = channels: "%.2f" % c[i].norm_kurtosis()),
            ("Entropy (from grey tone span)", lambda i, c = channels: "%.2f" % abs(c[i].entropy_g())),
            ("Entropy (from histogram)", lambda i, c = channels: "%.2f" % abs(c[i].entropy())),
            ("Entropy (from cooccurrence matrix)", lambda i, c = channels: "%.2f" % abs(c[i].entropy_c())),
            ("Anisotropycoefficient", lambda i, c = channels: "%.2f" % abs(c[i].anisotropycoefficient())),
            ("Contrast", lambda i, c = channels: "%.2f" % abs(c[i].contrast())),
            ("Homogenity", lambda i, c = channels: "%.2f" % c[i].homogenity()),
            ("RLE Effectivity", lambda i, c = channels: "%.2f" % c[i].rle_effectivity()))

        maximas = []
        minimas = []
        for i in channels:
            maximas.append(i.max())
            minimas.append(i.min())
            
        if max(maximas) <= 255 and min(minimas) >= 0:
            for i in range(len(computations)):
                Tkinter.Label(f, text = computations[i][0],
                              anchor = Tkinter.W).grid(row = i,
                                                       sticky = Tkinter.W)

            status = GUI.ProgressBar(self, labelText = "Calculating...")
            status.pack(padx=5, pady=5)

            self.update_idletasks()
            
            for i in range(len(computations)):
                for j in range(len(channels)):
                    Tkinter.Label(f, text = computations[i][1](j),
                                  anchor = Tkinter.W).grid(row = i,
                                                           column = j + 1,
                                                           sticky = Tkinter.E)
                    status.updateProgress(i * len(channels) + j + 1,
                                      len(channels) * len(computations))
                    self.update_idletasks()
            status.pack_forget()
        else:
            Tkinter.Label(f, text = "No more values available.\n" +
                          "This is not a quantized image.", anchor = Tkinter.W).pack()

        Tkinter.Button(self, text = "Close",
                       command = self.destroy,
                       default=Tkinter.ACTIVE).pack(padx=5, pady=5)

class HistogramWindow(ImageGUI.StaticImageWindow):
    "shows the histogram of an ImageWindow"
    def __init__(self, master):
        # save master image, "raw" histogram and (undefined) histogram image
        self.masterimage = master.getview()
        self.hist = Numeric.array(self.masterimage.histogram()) # list of 256 of 768 counts
        self.image = Image.new(self.masterimage.mode, (256,256))

        self.histmode = Tkinter.IntVar()
        self.histmode.set(0)    # 0 = normal, 1 = logarithmic, 2 = cumulative
        
        ImageGUI.StaticImageWindow.__init__(self, master, master.name + " Histogram")

        addon_structure = [
            ("separator",),
            ("radiobutton", "_Normal", self.changeviewmode, "", self.histmode, 0),
            ("radiobutton", "_Logarithmic", self.changeviewmode, "", self.histmode, 1),
            ("radiobutton", "_Cumulative", self.changeviewmode, "", self.histmode, 2)
            ]

        GUI.fillmenu(self.viewmenu, addon_structure, self)

    def getrawview(self):
        "override method of StaticImageWindow"

        mode = self.histmode.get()
        if mode == 0: # normal
            hist = self.hist.copy()
        elif mode == 1: # logarithmic
            hist = Numeric.log(self.hist + 1)
        elif mode == 2: # cumulative
            hist = []
            for i in range(len(self.hist) / 256): # channels
                sum = 0
                for j in range(256):
                    sum += self.hist[i * 256 + j]
                    hist.append(sum)
            hist = Numeric.array(hist)
        
        # normalize
        for i in range(len(hist) / 256):
            hist = hist * 256 / max(hist)
            
        for i in range(256):
            for j in range(256):
                if self.masterimage.mode == "RGB":
                    if j < 256 - hist[i]: r = 0
                    else: r = 255
                    if j < 256 - hist[i + 256]: g = 0
                    else: g = 255
                    if j < 256 - hist[i + 512]: b = 0
                    else: b = 255
                    colour = (r, g, b)
                else:
                    if j < 256 - hist[i]: colour = 0
                    else: colour = 255
                self.image.putpixel((i, j), colour)
            
        return self.image
        
    def getrawnumericimage(self):
        "override method of StaticImageWindow"
        self.image = self.getrawview()
        return ImageGUI.StaticImageWindow.getrawnumericimage(self)

class CooccurrenceWindow(ImageGUI.ColorImageWindow):
    "represents shows the cooccurence matrix in a window"
    def __init__(self, master):
        self.masterimage = master.getview()
        
        self.coocmode = Tkinter.IntVar()
        self.coocmode.set(0)    # 0 = horizontal, 1 = vertical, 2 = diagonal
        self.oldmode = -1
        
        ImageGUI.ColorImageWindow.__init__(self, master, master.name + " Cooccurrence Matrix ")

        addon_structure = [
            ("separator",),
            ("radiobutton", "_Horizontal", self.changeviewmode, "", self.coocmode, 0),
            ("radiobutton", "_Vertical", self.changeviewmode, "", self.coocmode, 1),
            ("radiobutton", "_Diagonal", self.changeviewmode, "", self.coocmode, 2)
            ]

        GUI.fillmenu(self.viewmenu, addon_structure, self)

    def getrawmode(self):
        "returns 'RGB' or 'L'"
        return self.masterimage.mode

    def getrawview(self):
        "implements for ImageWindow"
        mode = self.coocmode.get()
        if self.oldmode == mode: # cache
            return self.image
        if mode == 0:              # horizontal
            c = self.masterimage.size[0] - 1
            r = self.masterimage.size[1]
        elif mode == 1:            # vertical
            c = self.masterimage.size[0]
            r = self.masterimage.size[1] - 1
        elif mode == 2:            # diagonal
            c = self.masterimage.size[0] - 1
            r = self.masterimage.size[1] - 1
                
        if self.masterimage.mode == "L":
            C = Numeric.zeros((256, 256))
        else:
            CR = Numeric.zeros((256, 256))
            CG = Numeric.zeros((256, 256))
            CB = Numeric.zeros((256, 256))
        for i in range(c):
            for j in range(r):
                c1 = self.masterimage.getpixel((i, j))
                if mode == 0: # horizontal
                    c2 = self.masterimage.getpixel((i + 1, j))
                elif mode == 1: # vertical
                    c2 = self.masterimage.getpixel((i, j + 1))
                elif mode == 2: # diagonal
                    c2 = self.masterimage.getpixel((i + 1, j + 1))
                    
                if self.masterimage.mode == "L":
                    C[c1, c2] += 1
                else:
                    CR[c1[0], c2[0]] += 1
                    CG[c1[1], c2[1]] += 1
                    CB[c1[2], c2[2]] += 1

        if self.masterimage.mode == "L":
            C = C * 255 / max(max(C))
            self.image = NumericImage.NumImage(C).toImage()
        else:
            maximum = max(max(max(CR)), max(max(CG)), max(max(CB)))
            CR = CR * 255 / maximum
            CG = CG * 255 / maximum
            CB = CB * 255 / maximum
            self.image = Image.merge("RGB",
                                     (NumericImage.NumImage(CR).toImage(),
                                      NumericImage.NumImage(CG).toImage(),
                                      NumericImage.NumImage(CB).toImage()))
        self.oldmode = mode
        return self.image
        
    def getrawnumericimage(self):
        "implements for ColorImageWindow"
        image = self.getrawview()
        
        if image.mode == "L":
            return [NumericImage.NumImage(image)]
        else:
            return [
                NumericImage.NumImage(image.convert("L", (1, 0, 0, 0))),
                NumericImage.NumImage(image.convert("L", (0, 1, 0, 0))),
                NumericImage.NumImage(image.convert("L", (0, 0, 1, 0)))]
        
    def width(self):
        "implements for ImageWindow"
        return 256
        
    def height(self):
        "implements for ImageWindow"
        return 256

class RLMWindow(ImageGUI.StaticImageWindow):
    "represents shows the run length matrix in a window"
    def __init__(self, master):
        masterimage = master.getview()
        if masterimage.mode == "RGB":
            seq = list(masterimage.split())
        else:
            seq = [masterimage]
        result = []
        for i in range(len(seq)): # channels
            result.append(Numeric.zeros(
                (256, masterimage.size[0] + 1)).astype(Numeric.Float))
            for j in range(masterimage.size[1]): # rows
                rl = 0
                color = -1
                for k in range(masterimage.size[0]): # columns
                    oldcolor, color = color, seq[i].getpixel((k, j))
                    if color == oldcolor:
                        rl += 1
                    else:
                        if rl: result[i][oldcolor, rl] += rl
                        rl = 1
                if rl: result[i][color, rl] += rl
            #print Numeric.sum(Numeric.ravel(result[i]))
                # debug: should be 1 * imagesize

        maximum = max(max(max(result)))
        for i in range(len(result)):
            result[i] = result[i] * 255 / maximum
        if masterimage.mode == "RGB":
            self.image = Image.merge("RGB",
                            [NumericImage.NumImage(result[0]).toImage(),
                             NumericImage.NumImage(result[1]).toImage(),
                             NumericImage.NumImage(result[2]).toImage()])
        else: # "L"
            self.image = NumericImage.NumImage(result[0]).toImage()
            
        ImageGUI.StaticImageWindow.__init__(self, master, master.name + " RLM")

