#
# Copyright 2002 Roland Stigge
#
# This file is part of ImageWorkshop
#
# ImageWorkshop is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# ImageWorkshop is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with ImageWorkshop; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#

# system modules
import string

# Tkinter
import Tkinter
import tkMessageBox
import tkSimpleDialog
import tkFileDialog

# Python Imaging Library (PIL)
import ImageTk

# own modules
import GUI

class ScrollWindow(Tkinter.Toplevel):
    "represents and displays an image with scrollbars and zoom support"
    def __init__(self, master, name, photoimage):
        """
        constructor:
           master:      the parent widget (e.g. window)
           name:        the window title
           photoimage:  a ImageTk.PhotoImage, wraps a Tkinter.PhotoImage
        """
        self.name = name
        self.photoimage = photoimage
        Tkinter.Toplevel.__init__(self, master)
        self.protocol("WM_DELETE_WINDOW", self.exit)
        
        # create the scrollbars
        self.scrollbarx = Tkinter.Scrollbar(self, orient = Tkinter.HORIZONTAL)
        self.scrollbary = Tkinter.Scrollbar(self, orient = Tkinter.VERTICAL)

        # set initial window size
        self.scale = 1.0
        # scale down if any length bigger than half screen
        while (self.scaledimagewidth() > self.winfo_screenwidth() / 2 or
               self.scaledimageheight() > self.winfo_screenheight() / 2):
            self.scale /= 2
        # zoom if both width and height smaller than 32 pixels
        while (self.scaledimagewidth() < 32 and self.scaledimageheight() < 32):
            self.scale *= 2

        # set initial window size
        self.geometry("%dx%d" % (self.scaledimagewidth(),
                                 self.scaledimageheight()))

        # create the image canvas
        self.imagecanvas = Tkinter.Canvas(self,
                                          xscrollcommand = self.scrollbarx.set,
                                          yscrollcommand = self.scrollbary.set,
                                          highlightthickness = 0)
                                          # no 1 pixel border!
        self.imagecanvas.grid(row = 0, column = 0,
                              sticky = Tkinter.N+Tkinter.S+Tkinter.W+Tkinter.E)

        self.scrollbarx.config(command = self.imagecanvas.xview) # when slider
        self.scrollbary.config(command = self.imagecanvas.yview) #  is moved

        self.columnconfigure(0, weight=1)
        self.rowconfigure(0, weight=1)

        self.itemid = None    # init
        self.zoom(self.scale) # arrange window according to scale
        
        self.bind('<Configure>', self.adjustscrollbars)
        self.bind("+", self.zoomin)
        self.bind("-", self.zoomout)
        self.bind("<Alt-x>", self.exit)

        # update main window status bar
        self.bind("<Enter>", lambda event, self = self, text = self.name: self.updatestatus(event, text = text))
        self.bind("<Leave>", lambda event, self = self: self.updatestatus(event, text = ""))

        self.imagecanvas.bind("<Motion>", lambda event, self = self,
                              text = self.name: self.updatestatus(event,
                                                                  text = text,
            position = str(int(self.imagecanvas.canvasx(event.x) / self.scale))
              + ", " + str(int(self.imagecanvas.canvasy(event.y) / self.scale))
                                                                  ))

    def adjustscrollbars(self, event = None):
        """
        (Re-) Generate grid: with or without scrollbars
        to be called initially and on resize
        """
        if self.winfo_width() >= self.scaledimagewidth(): # no bottom scrollbar
            self.scrollbarx.grid_forget()
        else:                                             # bottom scrollbar
            self.scrollbarx.grid(row = 1, column = 0,
                                 sticky = Tkinter.W + Tkinter.E)
        if self.winfo_height() >= self.scaledimageheight():# no right scrollbar
            self.scrollbary.grid_forget()
        else:                                              # right scrollbar
            self.scrollbary.grid(row = 0, column = 1,
                                 sticky = Tkinter.N + Tkinter.S)
        
    def zoom(self, scale):
        """Apply a new zoom scale and rearrange window
        to be called initially and at zoom"""

        # set window title
        if scale == 1:
            zoomstring = ""
        else:
            zoomstring = "(" + str(int(scale*100)) + "%)"
        self.title(self.name + " " + zoomstring)

        # save old scrollbar (center) positions
        pos = self.scrollbarx.get()  # saved in tuple due to 4-/2-tuple "bug"
        avgx = (pos[0] + pos[1]) / 2
        pos = self.scrollbary.get()
        avgy = (pos[0] + pos[1]) / 2

        if avgx == 0 or avgy == 0:
            avgx = 0.5
            avgy = 0.5
        
        # new scale
        self.scale = scale
        if scale == 1: # maintain scale
            self.scaledimage = self.photoimage._PhotoImage__photo
        elif scale > 1: # zoom
            self.scaledimage = self.photoimage._PhotoImage__photo.zoom(int(scale))
        else:           # subsample
            self.scaledimage = self.photoimage._PhotoImage__photo.subsample(int(1/scale))
        # (re-)new item (with new scale)
        if self.itemid: # initial run
            self.imagecanvas.delete(self.itemid)
        self.itemid = self.imagecanvas.create_image(0, 0,
                                                    image = self.scaledimage,
                                                    anchor = Tkinter.NW)

        # set maximum window size
        self.maxsize(self.scaledimagewidth(), self.scaledimageheight())
        # -> shrink if too large
        if (self.winfo_width() > self.scaledimagewidth() or
            self.winfo_height() > self.scaledimageheight()):
            self.geometry("%dx%d" % (min(self.winfo_width(),
                                         self.scaledimagewidth()),
                                     min(self.winfo_height(),
                                         self.scaledimageheight())))

        # set scrollregion
        self.imagecanvas.config(scrollregion = (0, 0, self.scaledimagewidth(),
                                                self.scaledimageheight()))
        
        # restore relative scrollbar positions
        self.update_idletasks()
        lox, hix = self.scrollbarx.get()
        loy, hiy = self.scrollbary.get()

        radiusx = (hix - lox) / 2
        radiusy = (hiy - loy) / 2
        lox = avgx - radiusx
        hix = avgx + radiusx
        loy = avgy - radiusy
        hiy = avgy + radiusy
        # adjust if out of range
        if lox < 0:
            hix = hix - lox
            lox = 0
        if hix > 1:
            lox = lox - (hix - 1)
            hix = 1
        if loy < 0:
            hiy = hiy - loy
            loy = 0
        if hiy > 1:
            loy = loy - (hiy - 1)
            hiy = 1

        self.scrollbarx.set(lox, hix)
        self.scrollbary.set(loy, hiy)
        self.imagecanvas.xview(Tkinter.MOVETO, lox)
        self.imagecanvas.yview(Tkinter.MOVETO, loy)

        self.adjustscrollbars()

    def updateimage(self, photoimage):
        "replace the image with a new ImageTk.PhotoImage (with same size)"
        self.photoimage = photoimage
        if self.scale == 1: # maintain scale
            self.scaledimage = self.photoimage._PhotoImage__photo
        elif self.scale > 1: # zoom
            self.scaledimage = self.photoimage._PhotoImage__photo.zoom(int(self.scale))
        else:           # subsample
            self.scaledimage = self.photoimage._PhotoImage__photo.subsample(int(1/self.scale))
        self.imagecanvas.delete(self.itemid)
        self.itemid = self.imagecanvas.create_image(0, 0,
                                                    image = self.scaledimage,
                                                    anchor = Tkinter.NW)

    def scaledimagewidth(self):
        return int(self.photoimage._PhotoImage__photo.width() * self.scale)

    def scaledimageheight(self):
        return int(self.photoimage._PhotoImage__photo.height() * self.scale)

    def updatestatus(self, event = None, text = "", position = "   "):
        "update status bar in main window"
        ImageWorkshop.workshop.status.set(text)
        ImageWorkshop.workshop.position.set(position)

    def exit(self, event = None):
        "Close Image Window"
        self.destroy()

class ImageWindow(ScrollWindow):
    """
    represents and displays an image with menus and analysis and transform ops

    abstract class. deriving classes have to implement:
      getview() -> Image (viewmode dependent)
      getnumericimage() -> list of NumericImage.NumImage, represents view,
                           may be complex
      width() -> integer
      height() -> integer
    """
    
    def __init__(self, master, title = ""):
        """
        constructor:
           master:    the parent widget (e.g. Window)
           title:     the Window Title
        """
        ScrollWindow.__init__(self, master, title,
                              ImageTk.PhotoImage(self.getview()))

        ImageWorkshop.workshop.registerwindow(self)

        self.bind("<Destroy>", self.exit)

        # create the menu
        self.menu = Tkinter.Menu(self, title = "Image Menu")#, tearoff=0)

        menustructure = [
            ("cascade", "_File", "filemenu",
             [("command", "_Save...", self.save, ""),
              ("separator",),
              ("command", "_Close", self.exit, "<Alt-x>")
              ]),
            ("cascade", "_View", "viewmenu",
             [("command", "Zoom _In", self.zoomin, "+"),
              ("command", "Zoom _Out", self.zoomout, "-"),
              ("cascade", "_Zoom", "zoommenu",
               [("command", "16:1", self.zoom16_1, ""),
                ("command", "8:1", self.zoom8_1, ""),
                ("command", "4:1", self.zoom4_1, ""),
                ("command", "2:1", self.zoom2_1, ""),
                ("command", "1:1", self.zoom1_1, ""),
                ("command", "1:2", self.zoom1_2, ""),
                ("command", "1:4", self.zoom1_4, ""),
                ("command", "1:8", self.zoom1_8, ""),
                ("command", "1:16", self.zoom1_16, ""),
                ])
              ]),
            ("cascade", "Image _Data", "datamenu",
             [("command", "_Scalar Values", self.scalars, ""),
              ("command", "_Histogram", self.histogram, ""),
              ("command", "_Cooccurrence Matrix", self.cooccurrence, ""),
              ("command", "_Run Lenght Matrix", self.rlmatrix, "")
              ]),
            ("cascade", "_Point Operations", "pointmenu",
             [("command", "_Dynamic...", self.point_dynamic, ""),
              ("command", "_Threshold...", self.point_threshold, ""),
              ("command", "_Invert", self.point_invert, ""),
              ("command", "_Equalize (Histogram)", self.point_equalize, ""),
              ("command", "_Combine...", self.point_combine, ""),
              ("command", "_Salt and Pepper...", self.point_saltandpepper, ""),
              ("command", "_Bit Slice...", self.point_bitslice, ""),
              ]),
            ("cascade", "_Local Operations", "localmenu",
             [("command", "_Convolution Filters...", self.local_filter, ""),
              ("command", "_Rank Order Operations...", self.local_rankops, ""),
              ]),
            ("cascade", "_Global Operations", "globalmenu",
             [("command", "_Haar Transform", self.global_haar, ""),
              ("command", "_Walsh Transform, sequential",
               self.global_walsh_s, ""),
              ("command", "_Walsh Transform, binary",
               self.global_walsh_b, ""),
              ("command", "_Walsh Transform, natural",
               self.global_walsh_n, ""),
              ("separator",),
              ("command", "_Cosine Transform", self.global_cosine, ""),
              ("command", "_Fourier Transform", self.global_fourier, ""),
              ])
            ]

        GUI.fillmenu(self.menu, menustructure, self)

        self.bind("<Alt-m>", self.popup_menu)
        self.bind("<Button-3>", self.popup_menu)

    def popup_menu(self, event = None):
        "The popup menu callback"
        self.menu.tk_popup(event.x_root, event.y_root)

    def exit(self, event = None):
        "override ScrollWindow.exit()"
        ImageWorkshop.workshop.unregisterwindow(self)
        ScrollWindow.exit(self, event)

    #
    # File Menu entries
    #
    def save(self, event = None):
        "Save as File"
        filename = tkFileDialog.asksaveasfilename(filetypes = # write types
                    [('All files','*'),
                     ('Microsoft Windows BMP','*.bmp'),
                     ('Encapsulated Postscript', '*.eps'),
                     ('Graphics Interchange Format (GIF)','*.gif'),
                     ('Joint Photographic Expert Group (JPEG)','*.jpg'),
                     ('Joint Photographic Expert Group (JPEG)','*.jpe'),
                     ('Joint Photographic Expert Group (JPEG)','*.jpeg'),
                     ('PC Paintbrush (PCX)','*.pcx'),
                     ('Portable Document Format','*.pdf'),
                     ('Portable Network Graphics (PNG)','*.png'),
                     ('Portable Bitmap','*.pbm'),
                     ('Portable Bitmap','*.ppm'),
                     ('Portable Bitmap','*.pgm'),
                     ('Tag Image File Format (TIFF)','*.tiff'),
                     ('Tag Image File Format (TIFF)','*.tif')])
        if filename:
            if string.lower(filename[-4:]) in ('.jpg', '.jpe', 'jpeg'):
                sd = SaveDialogJPEG(self, title = "JPEG Save Options")
                if sd.result:
                    apply(self.getview().save,(filename,), sd.result)

            elif string.lower(filename[-4:]) == '.png':
                if tkMessageBox.askyesno('Save as PNG', 'Save in optimized mode?') == "yes":
                    self.getview().save(filename, optimize = 1)
                else:
                    self.getview().save(filename)
            else:
                self.getview().save(filename)
    #
    # View Menu entries
    #
    
    def zoomin(self, event = None):
        "Zoom 2x"
        self.scale *= 2
        self.zoom(self.scale)

    def zoomout(self, event = None):
        "Zoom out (x0.5)"
        self.scale /= 2
        self.zoom(self.scale)
        
    def zoom16_1(self, event = None):
        "Zoom 16:1"
        self.scale = 16.0
        self.zoom(self.scale)

    def zoom8_1(self, event = None):
        "Zoom 8:1"
        self.scale = 8.0
        self.zoom(self.scale)

    def zoom4_1(self, event = None):
        "Zoom 4:1"
        self.scale = 4.0
        self.zoom(self.scale)

    def zoom2_1(self, event = None):
        "Zoom 2:1"
        self.scale = 2.0
        self.zoom(self.scale)

    def zoom1_1(self, event = None):
        "Zoom 1:1"
        self.scale = 1.0
        self.zoom(self.scale)

    def zoom1_2(self, event = None):
        "Zoom 1:2"
        self.scale = 1.0 / 2
        self.zoom(self.scale)

    def zoom1_4(self, event = None):
        "Zoom 1:4"
        self.scale = 1.0 / 4
        self.zoom(self.scale)

    def zoom1_8(self, event = None):
        "Zoom 1:8"
        self.scale = 1.0 / 8
        self.zoom(self.scale)

    def zoom1_16(self, event = None):
        "Zoom 1:16"
        self.scale = 1.0 / 16
        self.zoom(self.scale)

    def changeviewmode(self, event = None):
        "Callback for view mode change"
        self.updateimage(ImageTk.PhotoImage(self.getview()))

    #
    # Image Data Menu entries
    #
    
    def scalars(self, event = None):
        "Show all Scalar Values"
        DataOps.ScalarsWindow(self)

    def histogram(self, event = None):
        "Show Histogram"
        DataOps.HistogramWindow(self)

    def cooccurrence(self, event = None):
        "Show Cooccurrence Matrix"
        DataOps.CooccurrenceWindow(self)

    def rlmatrix(self, event = None):
        "Show run length matrix"
        DataOps.RLMWindow(self)

    #
    # Point Operations Menu entries
    #

    def point_dynamic(self, event = None):
        "change dynamic (factor, offset)"
        d = PointOps.DynamicDialog(self, title = "Dynamic")
        if d.factor: # not canceled
            PointOps.DynamicImageWindow(self, d.factor, d.offset)

    def point_threshold(self, event = None):
        "apply threshold to make 'binary' image"
        threshold = tkSimpleDialog.askfloat("Threshold", "Please enter threshold:", initialvalue = 128, parent = self, minvalue = 0, maxvalue = 255)
        if threshold != None:
            PointOps.ThresholdImageWindow(self, threshold)

    def point_invert(self, event = None):
        "invert image"
        PointOps.InvertImageWindow(self)

    def point_equalize(self, event = None):
        "equalize (via histogram)"
        PointOps.EqualizeImageWindow(self)

    def point_combine(self, event = None):
        "combine 2 images"
        d = PointOps.CombineDialog(self, title = "Combine Images")
        if d.op != None: # not cancelled
            PointOps.CombineImageWindow(d.op, d.imagewindows)

    def point_saltandpepper(self, event = None):
        "add salt and pepper"
        d = PointOps.SandPDialog(self, title = "Salt and Pepper")
        if d.psalt:
            PointOps.SandPImageWindow(self, d.psalt, d.ppepper)

    def point_bitslice(self, event = None):
        "devide image into bit planes"
        d = PointOps.BitsliceDialog(self, title = "Bit slice")
        for i in d.planes:
            PointOps.BitsliceImageWindow(self, i)

    #
    # Local Operations Menu entries
    #

    def local_filter(self, event = None):
        "choose a filter and apply"
        d = LocalOps.FilterDialog(self, title = "Filter")
        if d.matrix != None:
            LocalOps.FilterImageWindow(self, d.matrix)

    def local_rankops(self, event = None):
        "median, erosion, dilatation, opening, closing filter"
        d = LocalOps.RankDialog(self, title = "Rank Order")
        if d.mode != None:
            LocalOps.RankImageWindow(self, d.mode, d.size)

    #
    # Local Operations Menu entries
    #

    def global_haar(self, event = None):
        "Haar Transform"
        GlobalOps.HaarImageWindow(self)

    def global_walsh_s(self, event = None):
        "Walsh Transform (sequential order)"
        GlobalOps.WalshImageWindow(self, order = "sequential")

    def global_walsh_b(self, event = None):
        "Walsh Transform (binary order)"
        GlobalOps.WalshImageWindow(self, order = "binary")

    def global_walsh_n(self, event = None):
        "Walsh Transform (natural order)"
        GlobalOps.WalshImageWindow(self, order = "natural")

    def global_fourier(self, event = None):
        "Fourier Transform"
        GlobalOps.FourierImageWindow(self)

    def global_cosine(self, event = None):
        "Cosine Transform"
        GlobalOps.CosineImageWindow(self)

class ColorImageWindow(ImageWindow):
    """
    An (abstract!) ImageWindow which can *potentially* represent a color image
    In this case, RGB, R, G and B mode (and for each, a grayscale mode)
      can be coosen.
    Implements getview() and getnumericimage().
    Instead, deriving classes must implement
      getrawmode() (-> "RGB" or "L" to get number of internal channels fast),
      getrawview() (-> (internal) Image, "RGB" or "L" mode) and
      getrawnumericimage() (-> list of NumericImage.NumImage instances)
    """

    def __init__(self, master, title = ""):
        """
        constructor:
           master:    the parent widget (e.g. Window)
           title:     the Window Title
        """
        self.viewmode = Tkinter.StringVar()
        self.viewmode.set("RGB")
        self.grayscale = Tkinter.IntVar()
        self.grayscale.set(0)
        
        ImageWindow.__init__(self, master, title = title)

        if self.getrawmode() == "RGB":
            addon_structure = [
                ("separator",),
                ("radiobutton", "_Color (RGB)", self.changeviewmode, "",
                 self.viewmode, "RGB"),
                ("radiobutton", "_Red Channel", self.changeviewmode, "",
                 self.viewmode, "R"),
                ("radiobutton", "_Green Channel", self.changeviewmode, "",
                 self.viewmode, "G"),
                ("radiobutton", "_Blue Channel", self.changeviewmode, "",
                 self.viewmode, "B"),
                ("separator",),
                ("checkbutton", "Gray_scale", self.changeviewmode, "",
                 self.grayscale)
                ]

            GUI.fillmenu(self.viewmenu, addon_structure, self)

    def getview(self):
        "return a view dependent Image object"
        raw = self.getrawview()
        if raw.mode == "L": # it's just an "L" (grayscale) file
            return raw
        elif not self.grayscale.get(): # color mode
            if self.viewmode.get() == "RGB":
                return raw
            elif self.viewmode.get() == "R":
                return raw.convert("RGB",(
                    1, 0, 0, 0,
                    0, 0, 0, 0,
                    0, 0, 0, 0))
            elif self.viewmode.get() == "G":
                return raw.convert("RGB",(
                    0, 0, 0, 0,
                    0, 1, 0, 0,
                    0, 0, 0, 0))
            else: # self.viewmode.get() == "B":
                return raw.convert("RGB",(
                    0, 0, 0, 0,
                    0, 0, 0, 0,
                    0, 0, 1, 0))
        else: # grayscale mode
            if self.viewmode.get() == "RGB":
                return raw.convert("L")
            elif self.viewmode.get() == "R":
                return raw.convert("L", (1, 0, 0, 0))
            elif self.viewmode.get() == "G":
                return raw.convert("L", (0, 1, 0, 0))
            else: # self.viewmode.get() == "B":
                return raw.convert("L", (0, 0, 1, 0))

    def getnumericimage(self):
        "return a view dependent list of NumericImage.NumImage objects"
        raw = self.getrawnumericimage()
        if len(raw) == 1: # it's just an "L" (grayscale) image
            return raw
        
        elif self.viewmode.get() == "R": # single colors ...
            return raw[0:1]
        elif self.viewmode.get() == "G":
            return raw[1:2]
        elif self.viewmode.get() == "B":
            return raw[2:3]
        elif self.grayscale.get(): # RGB -> Grayscale
            return [NumericImage.rgb2luma(raw)]
        else: # RGB tuple
            return raw

class StaticImageWindow(ColorImageWindow):
    """
    A ColorImageWindow where the image is represented by an PIL Image
    -> implements needed methods for ColorImageWindow
    -> deriving classes just have to implement an __init__()
       where self.image is constructed and call StaticImageWindow.__init__
    -> also useful for classes which will just override some of the derived
       methods, e.g. getrawimage()
    """

    def getrawmode(self):
        "return 'RGB' or 'L'"
        return self.image.mode

    def getrawview(self):
        "return the raw image (view)"
        return self.image
        
    def getrawnumericimage(self):
        "return the numeric version of the raw view"
        if self.image.mode == "L":
            return [NumericImage.NumImage(self.image)]
        else:
            return [
                NumericImage.NumImage(self.image.convert("L", (1, 0, 0, 0))),
                NumericImage.NumImage(self.image.convert("L", (0, 1, 0, 0))),
                NumericImage.NumImage(self.image.convert("L", (0, 0, 1, 0)))]
        
    def width(self):
        "the real width of this image"
        return self.image.size[0]

    def height(self):
        "the real height of this image"
        return self.image.size[1]
        
# own modules
import DataOps
import PointOps
import LocalOps
import GlobalOps
import NumericImage
import ImageWorkshop
