#
# Copyright 2002 Roland Stigge
#
# This file is part of ImageWorkshop
#
# ImageWorkshop is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# ImageWorkshop is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with ImageWorkshop; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#

#!/usr/bin/env python

import math, cmath
from types import *
import Image
import Numeric

class NumImage:
    """
    represents a (greyscale) Image with a Numeric.array
    using the Image module
    """
    #
    # Note: Python Imaging saves coordinates as (column, row)
    #       while we prefer (row, column)
    #
    
    #
    # Private data:
    #
    #   G = number of Colors
    #   M = Matrix
    #
        
    def __init__(self, image, G = 256):
        """
        creates an Image from an array, Image or File. e.g.:

          >>> ni = NumImage([[1,2],[3,4]], 8)  # a 2x2 3-bit image
            
          >>> im = Image.open('lena.jpg')
          >>> ni = NumImage(im)

          >>> im = Image.open('lena.jpg') # must be supported by Python Imaging

          G: Number of colors optional
        """

        if type(image) == ListType: # a list (of lists)
            self.G = G
            self.M = Numeric.array(image)
            
        elif type(image) == Numeric.arraytype: # an array
            self.G = G
            self.M = image

        elif isinstance(image, Image.Image): # an Image Object
            self.G = G
            #self.M = Numeric.zeros((image.size[1], image.size[0]))
            if image.format == "L":
                im = image
            else:
                im = image.convert("L")

            self.M = Numeric.array(im.getdata()).resize([im.size[1], im.size[0]])
            #for i in range(im.size[1]):
            #    for j in range(im.size[0]):
            #        self.M[i,j] = im.getpixel((j,i))
            
        elif isinstance(image, StringType): # a file(name)
            im = Image.open(image)

            self.G = G
            #self.M = Numeric.zeros((im.size[1], im.size[0]))
            if im.format != "L":
                im = im.convert("L")

            self.M = Numeric.array(im.getdata()).resize([im.size[1], im.size[0]])
            #for i in range(im.size[1]):
            #    for j in range(im.size[0]):
            #        self.M[i,j] = im.getpixel((j,i))

    def _get_G(self):
        "returns the number of colors"
        return self.G
        
    def _get_M(self):
        "returns the Matrix (as a Numeric.array)"
        return self.M

    def astype(self, t):
        "wraps astype() of Numeric.array"
        return NumImage(self._get_M().astype(t), self._get_G())
        
    def _set_M(self, M):
        "sets the Matrix (as a Numeric.array)"
        self.M = Numeric.around(M)
        
    def toImage(self):
        "returns the Python Image representation of this image"
        im = Image.new('L', (self.columns(), self.rows()), None)

        im.putdata(Numeric.resize(self.M, (self.columns() * self.rows(),)))
        
        #for i in range(self.rows()):
        #    for j in range(self.columns()):
        #        im.putpixel((j, i), self.M[i][j])
        return im
        
    def rows(self):
        "returns number of rows"
        return self.M.shape[0]
        
    def columns(self):
        "returns number of columns"
        return self.M.shape[1]

    def cornerstocenter(self):
        """
        returns the same Matrix, but with swapped quadrants:
            ( p q )  => ( d b )
            ( b d )     ( q p )
        (mostly used with frequency analysis)
        """

        u = self.M.shape[0] / 2
        v = self.M.shape[1] / 2

        M = Numeric.concatenate((Numeric.concatenate((self.M[u:,v:], self.M[:u,v:]),0),Numeric.concatenate((self.M[u:,:v], self.M[:u,:v]),0)), 1)
        
        return NumImage(M, self.G)

    #
    # computation some image data values
    #

    def numberofdifferentvalues(self):
        "returns the number of different gray tone values"
        dict = {}
        a = self.M.tolist()
        for i in range(self.M.shape[0]):
            for j in range(self.M.shape[1]):
                dict.setdefault(a[i][j], 1)
        return len(dict)

    def pixels(self):
        "number of pixel in whole image"
        return self.M.shape[0] * self.M.shape[1]
    
    def max(self):
        "returns the maximum value in this image"
        return max(Numeric.ravel(self.M))
    
    def min(self):
        "returns the minimum value in this image"
        return min(Numeric.ravel(self.M))

    def grayspan(self):
        "returns the gray tone span"
        return self.max() - self.min()

    def dynamic(self):
        "returns dynamic (global contrast)"
        max = float(self.max())
        min = float(self.min())
        if max + min == 0:
            return 0 # "undefined"
        return (max - min) / (max + min)

    def modal(self):
        "Value which occurs most often"
        list =  Numeric.ravel(self.M).tolist()
        list.sort()
        current = list[0] - 1
        list.append(list[len(list) - 1] + 1)
        number = 0          # number of equal elements in order
        mediannumber = 0
        for i in list:
            if current != i: # different/new value found
                if number > mediannumber:
                    mediannumber = number
                    median = current
                current = i
                number = 1
            else:
                number += 1
        return median

    def median(self):
        "Value in the middle of sorted list of gray values"
        temp = Numeric.ravel(self.M).tolist()
        temp.sort()
        return temp[len(temp)/2]

    def average(self):
        "returns mean value"
        return Numeric.average(Numeric.ravel(self.M))

    def weighted_moment(self, n):
        "returns n-th weighted moment"
        hist = self.normhistdict()
        avg = self.average()
        sum = 0.0
        list = Numeric.ravel(self.M).tolist()
        for k, v in hist.items():
            sum += (k - avg) ** n * v
        return sum

    def norm_wm(self, n):
        "returns normalized n-th weighted moment"
        if self.weighted_moment(2) ** (n / 2.0) == 0:
            return 0 # "undefined"
        return (self.weighted_moment(n) /
                self.weighted_moment(2) ** (n / 2.0))
    
    def variance(self):
        "returns variance / dispersion"
        return self.weighted_moment(2)
    
    def stddev(self):
        "returns standard deviation"
        return math.sqrt(self.variance())
    
    def asymmetry(self):
        "returns asymmetry"
        return self.weighted_moment(3)
    
    def norm_asymmetry(self):
        "returns normalized asymmetry"
        return self.norm_wm(3)
    
    def kurtosis(self):
        "returns kurtosis / excess"
        return self.weighted_moment(4)
    
    def norm_kurtosis(self):
        "returns normalized kurtosis / excess"
        return self.norm_wm(4)
    
    def entropy_g(self):
        "returns entropy calculated from gray tone span"
        return math.log(self.grayspan() + 1) / math.log(2)

    def normhistdict(self):
        "returns normalized histogram as a dictionary"
        dict = {}
        a = self.M.tolist()
        # sum up
        for i in range(self.M.shape[0]):
            for j in range(self.M.shape[1]):
                dict[a[i][j]] = dict.get(a[i][j], 0) + 1
        # normalize
        size = float(self.pixels())
        for i in dict.keys():
            dict[i] /= size
        return dict

    def entropy(self):
        "returns entropy calculated from the histogram"
        hist = self.normhistdict()
        sum = 0.0
        for i in hist.values():
            sum -= i * math.log(i) / math.log(2)
        return sum

    def anisotropycoefficient(self):
        "returns anisotropy coefficient"
        median = self.median()
        hist = self.normhistdict()
        sum1 = sum2 = 0.0
        for k, v in hist.items():
            x = - v * math.log(v) / math.log(2)
            if k <= median:
                sum1 += x
            sum2 += x
        if sum2 == 0:
            return 0 # "undefined"
        return sum1 / sum2

    def cooc(self):
        "returns cooccurrence matrix (G x G)"
        image = Numeric.minimum(Numeric.maximum(
            self.M.astype(Numeric.Int), 0), self.G - 1)
        C = Numeric.zeros((self.G, self.G)).astype(Numeric.Float)
        for i in range(self.M.shape[0]):
            for j in range(self.M.shape[1] - 1):
                C[int(self.M[i,j]), int(self.M[i,j + 1])] += 1
        C /= self.M.shape[0] * (self.M.shape[1] - 1)
        #print Numeric.sum(Numeric.ravel(C)) # debug: must be 1
        return C
        
    def entropy_c(self):
        "returns entropy calculated from cooccurrence matrix"
        C = self.cooc()
        sum = 0.0
        for i in range(self.G):
            for j in range(self.G):
                C_element = C[i,j]
                if C_element:
                    sum -= C_element * math.log(C_element) / math.log(2)
        return sum / 2 # because 2-overlapping

    def contrast(self):
        "returns contrast (calculated from cooccurrence matrix)"
        C = self.cooc()
        sum = 0.0
        for i in range(self.G):
            for j in range(self.G):
                # exponent 0.25 is "random"!
                sum += (float(abs(i - j)) / self.G) ** 0.25 * C[i,j]
        return sum

    def homogenity(self):
        "returns homogenity (calculated from cooccurrence matrix)"
        return 1 - self.contrast()

    def rlmatrix(self):
        "returns the run length matrix (rows = colors, columns = run lengths)"
        image = Numeric.minimum(Numeric.maximum(
            self.M.astype(Numeric.Int), 0), self.G - 1)
        RLM = Numeric.zeros((self.G, self.M.shape[1] + 1)).astype(Numeric.Float)
        # get runs
        for i in range(self.M.shape[0]): # rows
            current = self.M[i, 0] + 1 # current gray tone
            rl = 0
            for j in range(self.M.shape[1]): # columns
                if self.M[i, j] != current:
                    if rl: RLM[int(self.M[i, j-1]), rl] += rl
                    rl = 1
                    current = self.M[i, j]
                else:
                    rl += 1
            if rl: RLM[int(self.M[i, j]), rl] += rl
        RLM /= self.pixels()
        #print Numeric.sum(Numeric.ravel(RLM)) # debug: must be 1
        return RLM
        
    def rle_effectivity(self):
        "returns effectivity of an RL encoding"
        RLM = self.rlmatrix()
        sum = 0.0
        for g in range(self.G):
            for l in range(self.M.shape[1] + 1):
                sum += (float(l) / self.M.shape[1]) ** 2 * RLM[g, l]
                    # exp. 2 is random!
        return sum
        
def rgb2luma(list):
    """
    returns a 'grayscale' NumImage from a (R,G,B) list of NumericImage
    instances using ITU-R 601-2 luma transform:
       result = R * 299/1000 + G * 587/1000 + B * 114/1000
    """
    return NumImage((299.0 * list[0]._get_M() +
                     587.0 * list[1]._get_M() +
                     114.0 * list[2]._get_M())/1000)

def getbit(x, n):
    "returns the n-th bit (2^0-based) of the number x"
    return ((2 ** n) & x) != 0

def walshmatrix(N, order):
    """
    returns a (square) walsh matrix
    (intended for walsh transform, DWT)
    N = size of the matrix (NxN)
    order = "sequential" or "binary" or "natural"
    """

    n = int(math.log(N) / math.log(2))
    result = Numeric.zeros((N,N), Numeric.Float)
    if order == "sequential":
        for u in range(N):
            for v in range(N):
                sum = 0
                for z in range(n):
                    sum += getbit(v, n-1-z) * (getbit(u, z+1) + getbit(u, z))
                result[u,v] = (-1) ** sum

    elif order == "binary":
        for u in range(N):
            for v in range(N):
                sum = 0
                for z in range(n):
                    sum += getbit(v, z) * getbit(u, n-z-1)
                result[u,v] = (-1) ** sum

    elif order == "natural":
        for u in range(N):
            for v in range(N):
                sum = 0
                for z in range(n):
                    sum += getbit(v, z) * getbit(u, z)
                result[u,v] = (-1) ** sum
                
    else:
        raise "Order (" + order + ") not valid."

    return result / math.sqrt(N)

def cosinematrix(N):
    "returns a cosinematrix (intended for cosine transform, DCT)"

    result = Numeric.zeros((N,N), Numeric.Float)
    for u in range(N):
        for v in range(N):
            if u == 0:
                result[u,v] = math.sqrt(1.0/N)
            else:
                result[u,v] = math.sqrt(2.0/N) * math.cos((2.0*v+1)*u*math.pi/(2.0*N))
    return result

def haarmatrix(N):
    "returns a haarmatrix (intended for haar transform, DHT)"

    result = Numeric.zeros((N,N), Numeric.Float)
    
    value = 1.0           # value of matrix element
    seed = 1              # init for mother - wavelet
    v = 0                 # columns
    step = 2 * N
    for u in range(N):    # rows
        counter = 0
        
        while counter < step / 2:
            result[u, v % N] = value
            counter += 1
            v += 1
            
        while counter < step and counter < N:
            result[u, v % N] = -value
            counter += 1
            v += 1
            
        if (v % N) == 0:
            if u != 0:
                seed *= 2
                value = math.sqrt(seed)
            step /= 2

    return result / math.sqrt(N)

def fouriermatrix(N):
    "returns a matrix intended for fourier transform (DFT)"

    result = Numeric.zeros((N,N), Numeric.Complex)

    for u in range(N):
        for v in range(N):
            result[u,v] = cmath.exp(-2.0j*cmath.pi*u*v/N)

    return result / cmath.sqrt(N)

