#
# Copyright 2002 Roland Stigge
#
# This file is part of ImageWorkshop
#
# ImageWorkshop is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# ImageWorkshop is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with ImageWorkshop; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#

# system modules
import random

# Tkinter
import Tkinter
import tkMessageBox

# Python IMaging Library (PIL)
import Image

# Numerical Python
import Numeric

# own modules
import GUI
import ImageGUI
import NumericImage
import ImageWorkshop

#
# classes for point operations
#

class InvertImageWindow(ImageGUI.StaticImageWindow):
    "inverts an image"
    def __init__(self, master):
        self.image = master.getview().point(lambda i: 255 - i)

        ImageGUI.StaticImageWindow.__init__(self, master, master.name + " inverted")
        
class DynamicDialog(GUI.Dialog):
    """
    asks for factor and offset for dynamic correction
    return values in: self.factor, self.offset
    if successful: self.factor != None
    """

    def body(self, master):
        Tkinter.Label(master, text = "Factor:").grid(row = 0)
        Tkinter.Label(master, text = "Offset:").grid(row = 1)

        self.factorentry = Tkinter.Entry(master)
        self.factorentry.insert(Tkinter.END, "1.0")
        self.offsetentry = Tkinter.Entry(master)
        self.offsetentry.insert(Tkinter.END, "0.0")
        
        self.factorentry.grid(row = 0, column = 1)
        self.offsetentry.grid(row = 1, column = 1)

        self.factor = None

        return self.factorentry

    def validate(self):
        try:
            self.factor = float(self.factorentry.get())
            self.offset = float(self.offsetentry.get())
        except ValueError:
            tkMessageBox.showwarning("Bad input", "Values must be float")
            self.factor = None
            return 0

        return 1

class DynamicImageWindow(ImageGUI.StaticImageWindow):
    def __init__(self, master, factor, offset):
        self.image = master.getview().point(lambda i, factor = factor, offset = offset: i * factor + offset)

        ImageGUI.StaticImageWindow.__init__(self, master,
                                            master.name + " [dynamic changed]")
        
class ThresholdImageWindow(ImageGUI.StaticImageWindow):
    def __init__(self, master, threshold):
        self.image = master.getview().point(lambda i, threshold = threshold: (i >= threshold) * 255)

        ImageGUI.StaticImageWindow.__init__(self, master,
                                            master.name + " [threshold]")
        
class SandPDialog(GUI.Dialog):
    """
    asks for salt and pepper probability
    return values in: self.psalt, self.ppepper
    if successful: self.psalt != None
    """

    def body(self, master):
        Tkinter.Label(master, text = "Please enter probabilities:").grid(row = 0, columnspan = 2, sticky = Tkinter.W)
        Tkinter.Label(master, text = "Salt:").grid(row = 1)
        Tkinter.Label(master, text = "Pepper:").grid(row = 2)

        self.psaltentry = Tkinter.Entry(master)
        self.psaltentry.insert(Tkinter.END, "0.001")
        self.ppepperentry = Tkinter.Entry(master)
        self.ppepperentry.insert(Tkinter.END, "0.001")
        
        self.psaltentry.grid(row = 1, column = 1)
        self.ppepperentry.grid(row = 2, column = 1)

        self.psalt = None

        return self.psaltentry

    def validate(self):
        try:
            self.psalt = float(self.psaltentry.get())
            self.ppepper = float(self.ppepperentry.get())
        except ValueError:
            tkMessageBox.showwarning("Bad input", "Values must be float")
            self.psalt = None
            return 0

        return 1

class SandPImageWindow(ImageGUI.StaticImageWindow):
    def __init__(self, master, psalt, ppepper):
        self.image = master.getview().copy()
        if self.image == "L":
            color = lambda i: i
        else:
            color = lambda i: (i, i, i)
        # salt == white
        for i in range(int(master.width() * master.height() * psalt)):
            self.image.putpixel((random.randrange(master.width()),
                                 random.randrange(master.height())), color(255))
        # pepper == black
        for i in range(int(master.width() * master.height() * ppepper)):
            self.image.putpixel((random.randrange(master.width()),
                                 random.randrange(master.height())), color(0))

        ImageGUI.StaticImageWindow.__init__(self, master,
                                            master.name + " [salt'n'pepper]")

class EqualizeImageWindow(ImageGUI.StaticImageWindow):
    def __init__(self, master):
        image = master.getview()
        # get histogram
        if image.mode == "L":
            hist = image.histogram()
        else: # "RGB"
            grey = image.convert("L")
            hist = grey.histogram()

        # calculate LUT
        LUT = []
        sum = 0
        for i in range(256): # create LUT
            sum += hist[i]
            LUT.append(sum)
        LUT = Numeric.array(LUT)
        LUT = LUT * 255 / (image.size[0] * image.size[1])

        # Equalize
        if image.mode == "L":
            self.image = image.point(lambda i, LUT = LUT: LUT[i])
        else: # "RGB"
            # create 3 NumImages
            n = list(image.split())
            for c in range(3): # for all channels
                n[c] = NumericImage.NumImage(n[c])

                for i in range(n[c].rows()):
                    for j in range(n[c].columns()):
                        greypixel = grey.getpixel((j, i))
                        n[c].M[i, j] = n[c].M[i, j] * float(LUT[greypixel]) / greypixel
            
            minimum = min(map(lambda i, n = n: min(Numeric.ravel(n[i].M)), range(3)))
            for c in range(3):
                n[c].M = n[c].M - minimum
            maximum = max(map(lambda i, n = n: max(Numeric.ravel(n[i].M)), range(3)))
            if maximum:
                for c in range(3):
                    n[c].M = n[c].M * 255 / maximum
            for c in range(3):
                n[c] = n[c].toImage()
            self.image = Image.merge("RGB", n)
                                
        ImageGUI.StaticImageWindow.__init__(self, master,
                                            master.name + " [equalized]")

COMBINEOPS = ["Addition / Averaging", "Subtraction", "XOR", "OR", "AND"]

class CombineDialog(GUI.Dialog):
    """
    asks for images (ImageWindows) to combine and the operation to perform
    return values in: self.op, self.imagewindows
    if successful: self.op != None
    """

    def body(self, master):
        labels = ["First operand:", "Second operand:", "Operation:"]
        self.listboxes = [] # "raw" results
        for i in range(3):
            Tkinter.Label(master, text = labels[i]).grid(row = 0, column = i)
            
            frame = Tkinter.Frame(master)
            frame.grid(row = 1, column = i)

            scrollbar = Tkinter.Scrollbar(frame, orient = Tkinter.VERTICAL)
            listbox = Tkinter.Listbox(frame, exportselection = 0,
                                      yscrollcommand = scrollbar.set)
            scrollbar.config(command = listbox.yview)
            scrollbar.pack(side = Tkinter.RIGHT, fill = Tkinter.Y)
            listbox.pack(side = Tkinter.LEFT, fill = Tkinter.BOTH, expand = 1)
            self.listboxes.append(listbox)

            if i < 2: # operands
                listbox.config(width = 40)
                for j in ImageWorkshop.workshop.registeredwindows():
                    listbox.insert(Tkinter.END, j.name)
            else: # operator
                for j in COMBINEOPS:
                    listbox.insert(Tkinter.END, j)
                    
            if listbox.size() > 0: # select first one if present
                listbox.select_set(0)

        self.op = None

        return self.listboxes[0]

    def validate(self):
        windowlist = ImageWorkshop.workshop.registeredwindows()
        selections = map(lambda i: i.curselection(), self.listboxes)
        try:
            selections = map(lambda i: map(int, i), selections)
        except ValueError: pass

        self.imagewindows = map(lambda i, windowlist = windowlist:
                                windowlist[i[0]],
                                selections[0:2])
        self.op = selections[2][0]

        return 1

class CombineImageWindow(ImageGUI.StaticImageWindow):
    def __init__(self, op, imagewindows):
        im1 = imagewindows[0].getnumericimage()
        im2 = imagewindows[1].getnumericimage()

        if im1[0].rows() != im2[0].rows() or im1[0].columns() != im2[0].columns():
            tkMessageBox.showerror("Error", "Images do not have the same size")
            return 0
        
        if len(im1) < len(im2): # im1 "L", im2 "RGB"
            im1.extend([im1[0], im1[0]])
        elif len(im2) < len(im1): # im2 "L", im1 "RGB"
            im2.extend([im2[0], im2[0]])

        result = []
        for i in range(len(im1)):
            if op == 0: # Averaging
                result.append((im1[i]._get_M() + im2[i]._get_M()) / 2)
            elif op == 1: # Subtraction
                result.append((im1[i]._get_M() - im2[i]._get_M()) / 2 + 128)
            elif op == 2: # XOR
                result.append(im1[i]._get_M().astype(Numeric.Int) ^
                              im2[i]._get_M().astype(Numeric.Int))
            elif op == 3: # OR
                result.append(im1[i]._get_M().astype(Numeric.Int) |
                              im2[i]._get_M().astype(Numeric.Int))
            else: # op == 4: # AND
                result.append(im1[i]._get_M().astype(Numeric.Int) &
                              im2[i]._get_M().astype(Numeric.Int))

        if len(result) == 1: # "L"
            self.image = NumericImage.NumImage(result[0]).toImage()
        else: # "RGB"
            self.image = Image.merge("RGB", [
                NumericImage.NumImage(result[0]).toImage(),
                NumericImage.NumImage(result[1]).toImage(),
                NumericImage.NumImage(result[2]).toImage()])

        name = COMBINEOPS[op] + "("
        for i in imagewindows:
            name += i.name + ", "
        name = name[:-2] + ")"
        
        ImageGUI.StaticImageWindow.__init__(self, ImageWorkshop.workshop, name)


class BitsliceDialog(GUI.Dialog):
    """
    asks for planes to display
    return value in: self.planes (list)
    """

    def body(self, master):
        Tkinter.Label(master, text = "Choose which bit plane to display:").pack()
        self.vars = []
        buttons = []
        for i in range(8):
            var = Tkinter.IntVar()
            var.set(1)
            buttons.append(Tkinter.Checkbutton(master, text = "Plane " +
                                               str(i), variable = var))
            buttons[-1].pack()
            self.vars.append(var)
            
        self.planes = []

        return buttons[0]

    def apply(self):
        for i in range(8):
            if self.vars[i].get():
                self.planes.append(i)

class BitsliceImageWindow(ImageGUI.StaticImageWindow):
    def __init__(self, master, plane):
        self.image = master.getview().point(lambda i, plane = plane:
                                            ((i & (2 ** plane)) != 0) * 255)

        ImageGUI.StaticImageWindow.__init__(self, master, master.name +
                                            " [Bit plane " + str(plane) + "]")
        
